from typing import Dict

from django.db.models import ManyToOneRel, ForeignKey
from django.db.models import OneToOneRel
from django.db.models import ManyToManyRel
from django.db.models import ForeignObjectRel

REVERSE_RELATIONS = [ManyToOneRel, OneToOneRel, ManyToManyRel, ForeignObjectRel]


class BaseModelFieldsMixin:

    @classmethod
    def get_model_fields(cls, model) -> Dict:
        return {field.name: field for field in model._meta.get_fields() if not cls.is_reverse_relation_field(field)}

    @classmethod
    def is_reverse_relation_field(cls, field):
        for rel_relation in REVERSE_RELATIONS:
            if isinstance(field, rel_relation):
                return True
        return False

    @classmethod
    def get_model_raw_fields(cls, model):
        fields = cls.get_model_fields(model)
        return {
            cls.get_raw_field(field): field for field in fields.values()
        }

    @classmethod
    def get_raw_field(cls, field):
        if cls.is_foreign_key(field):
            return f'{field.name}_id'

        return field.name

    @classmethod
    def is_foreign_key(cls, cls_field):
        return isinstance(cls_field, ForeignKey)

    @classmethod
    def get_non_id_fields(cls, model):
        fields = [
            field for field
            in cls.get_model_fields(model).keys()
            if 'id' != field
        ]
        return fields



class PandasORMError(Exception):
    pass


class DataFrameSaveError(PandasORMError):

    def __init__(self, error=None, msg=None):
        self.error = error
        self.msg = msg
        super(__class__, self).__init__(msg)


class DataFrameModelNotSpecified(PandasORMError):
    pass



from typing import List

from pandas import DataFrame

from pandas_orm.django.mixins.base_model_fields_mixin import BaseModelFieldsMixin


class DjangoDataFrameMixin:
    model_mixin = BaseModelFieldsMixin

    @classmethod
    def add_dataframe_ids(cls, df: DataFrame, saved):
        if isinstance(saved, int):
            return df
        ids = [obj.pk for obj in saved]
        df['id'] = ids
        return df

    @classmethod
    def get_objs(cls, df, model):
        report_records = [
            cls.get_object_model(model, **record) for record in df.to_dict('records')
        ]
        return report_records

    @classmethod
    def get_object_model(cls, model, id=None, pk=None, **kwargs):
        obj = model(**kwargs)
        obj.pk = id or pk
        return obj

    @classmethod
    def clean_records_columns(cls, df, model):
        fields = DjangoDataFrameMixin.get_model_fields(model)
        df = cls.fill_missing_columns(df, fields)

        columns = [col for col in df.columns if
                   col in col in fields.keys() and col != 'id']
        df_result = df[columns]
        return df_result

    @classmethod
    def get_model_fields(cls, model):
        return cls.model_mixin.get_model_raw_fields(model)

    @classmethod
    def fill_missing_columns(cls, df, fields: List[str]):
        fields_missing = [
            col for col, field_cls in fields.items()
            if
            col not in df.columns and not cls.is_relation_id_field(field_cls)
        ]
        for missing_field in fields_missing:
            df[missing_field] = None

        return df

    @classmethod
    def is_relation_id_field(cls, field_cls):
        return cls.model_mixin.is_foreign_key(field_cls) or \
               cls.model_mixin.is_reverse_relation_field(field_cls) or \
               field_cls.name == 'id'



from contextlib import contextmanager
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine


databases = dict()


def get_global_instance_db(db_url):
    """
        modify global copy of databases for global instances usage
        to create only one instance per python interpreter usage
    :param db_url: str
    :return: DatabaseSession
    """
    global databases
    if db_url not in databases:
        databases[db_url] = DatabaseSession(db_url)
    return databases[db_url]


class DatabaseSession:
    """
    responsible to handle engine scope and session scope with sqlalchemy frameworke
    Usage:
    In [0]: db = DatabaseSession(url)
    In [1]: with db.session() as session:
       ...:     result = session.execute('select * from collaborator where id=144')
    In [2]: result.fetchall()
    Out[2]: [(1, u'name name', u'test test',...]
    """
    def __init__(self, url=None, session_scope_func=None, engine_scope_func=None):
        """
        :param url: url for the database
        :param session_scope_func: or session_scope to initialize sqlalchemy session
        :param engine_scope_func: or engine_scope to initialize sqlalchemy engine
        """
        self.url = url
        self.session_scope_func = session_scope_func
        self.engine_scope_func = engine_scope_func

    def session(self):
        """
        :return: session scope function context
        """
        if self.session_scope_func:
            return self.session_scope_func()
        return session_scope(self.url)

    def engine(self):
        """
        :return: engine scope function context
        """
        if self.engine_scope_func:
            return self.engine_scope_func()
        engine = engine_scope(self.url)
        return engine

    @classmethod
    def get_db_session(cls, url):
        """ initialize global variable singleton """
        return get_global_instance_db(url)


@contextmanager
def session_scope(db_url):
    """Provide a transactional scope around a series of operations."""
    engine, session = get_session_engine(db_url)
    try:
        yield session
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()
        engine.dispose()


@contextmanager
def engine_scope(db_url):
    """Provide a transactional scope around a series of operations."""
    engine, session = get_session_engine(db_url)
    try:
        yield engine
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()
        engine.dispose()


def get_session_engine(db_url):
    """
    configure Session class with desired options
    :param db_url: str connection for sqlalchemy_backend
    :return: engine, Session
    """
    engine = create_engine(db_url)
    Session = sessionmaker(bind=engine)
    conn = engine.connect()
    session = Session(bind=conn)
    return engine, session



from sqlalchemy.engine.url import make_url


def get_django_db(url, engine='django.db.backends.postgresql'):
    """
    :param url: Database string representation
    :param engine: database engine for django framework
    :return: Django Dict Database Representation
    """
    db_url = get_db_url(url)
    database_setup = {
        'ENGINE': engine,
        'NAME': db_url.database,
        'USER': db_url.username,
        'PASSWORD': db_url.password,
        'HOST': db_url.host,
        'PORT': db_url.port,
    }
    return database_setup


def get_db_url(url):
    return make_url(url)



import logging

from pandas import DataFrame

from pandas_orm.sqlalchemy.execute_sql import DataFrameExecuteSql
from pandas_orm.sqlalchemy.crud.naive_save_arguments import NativeModelSaveArguments


class ModelDataFrameManager:

    @classmethod
    def bulk_save(cls, records, model, engine_context_func, unique_fields=None,
                   update_fields=None, returning_id=False):
        """
        :param records: DataFrame or collection
        :param model: sqlalchemy.orm.declarative_base
        :param engine_context_func: engine_scope context
        :param unique_fields: List[field..]
        :param update_fields: List[field..]
        :param returning_id: bool
        :return: saved DataFrame
        """
        records = cls._prepare_records(records)
        if records.empty:
            return records
        logging.info(f'Saving {records.shape[0]} {model.__name__} records')
        kwargs = NativeModelSaveArguments(
            records=records,
            model=model,
            unique_fields=unique_fields,
            update_fields=update_fields).get_args()
        with engine_context_func() as engine:
            kwargs['engine'] = engine
            new_records = records
            if returning_id:
                results = DataFrameExecuteSql.save_dataframe_to_sql_returning_id(
                    **kwargs)
                new_records = cls._add_records_ids(results, records)
            else:
                DataFrameExecuteSql.save_dataframe_to_sql(**kwargs)

            return new_records

    @classmethod
    def _add_records_ids(cls, results, records):
        new_records = records
        new_ids = results.fetchall()
        if issubclass(records.__class__, DataFrame):
            df_ids = DataFrame(new_ids, columns=['id'])
            records['id'] = df_ids['id']
            new_records = records.to_dict('records')
        if isinstance(records, list):
            [record.update({'id': new_id[0]}) for record, new_id in
             zip(new_records, new_ids)]
        return new_records

    @classmethod
    def _prepare_records(cls, records) -> DataFrame:
        if issubclass(records.__class__, DataFrame):
            return records
        if isinstance(records, list):
            return DataFrame(records)

        raise NotImplementedError(type(records))





import time
from functools import wraps

from django.db.models import QuerySet

from pandas_orm.base.log import get_logger
from pandas_orm.django.dataframe import DataFrame


def get_result_query_columns(result_query: QuerySet):
    """
    :param result_query: django.db.models.QuerySet
    :return: List[field_name] of the django model
    """
    return [field.name for field in result_query.model._meta.fields]


def query_to_dataframe(query: QuerySet):
    """
    :param query: django.db.models.QuerySet
    :return: DataFrame
    """
    logger = get_logger()
    columns = get_result_query_columns(query)
    if not query:
        df = DataFrame([], columns=columns)
        df.__model__ = query.model
        return df
    time_started = time.time()
    logger.debug('Querying started %s ...' % query.query)
    list_values = query.values_list(*columns)
    result_dataframe = DataFrame(list(list_values), columns=columns)
    result_dataframe.__model__ = query.model
    querying_time = time.time() - time_started
    logger.debug('Done Querying in %f seconds ' % querying_time)
    return result_dataframe


def to_dataframe(func):
    """
    Annotation that wraps django.db.models.QuerySet reads to a DataFrame
    :param func: function that returns django.db.models.QuerySet
    :return: DataFrame
    """
    @wraps(func)
    def wrapper(*args, **kwargs) -> DataFrame:
        result_query = func(*args, **kwargs)
        return query_to_dataframe(result_query)

    return wrapper


def django_dataframe_values(func):
    """
    Annotation that wraps django.db.models.QuerySet reads to a DataFrame
    as values attributes
    :param func: function that returns django.db.models.QuerySet
    :return: DataFrame
    """
    @wraps(func)
    def wrapper(*args, **kwargs) -> DataFrame:
        result_query = func(*args, **kwargs)
        logger = get_logger()
        columns = get_result_query_columns(result_query)
        if not result_query:
            return DataFrame([], columns=columns)
        time_started = time.time()
        logger.debug('Querying started %s ...' % result_query.query)
        result_dataframe = DataFrame(list(result_query))
        querying_time = time.time() - time_started
        logger.debug('Done Querying in %f seconds ' % querying_time)
        return result_dataframe

    return wrapper



from abc import abstractmethod

from pandas import DataFrame as PDDataFrame

from pandas_orm.base.exceptions import DataFrameModelNotSpecified


class BaseDataFrame(PDDataFrame):

    def __init__(self, *args, orm_model=None, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.__model__ = orm_model

    @property
    def model(self):
        if self.__model__ is None:
            raise DataFrameModelNotSpecified()
        return self.__model__

    @abstractmethod
    def to_objects(self, *args, **kwargs):
        raise NotImplementedError()



import time
from functools import wraps

from pandas import DataFrame

from pandas_orm.base.log import get_logger
from pandas_orm.sqlalchemy.exceptions import AlchemyUpdateColumnsNotSupported

from sqlalchemy.dialects.postgresql import insert


def execute_sql(func):
    """
    Reads the sql alchemy statement queryset as a dataframe
    :param func: function that returns statement, engine
    :return: DataFrame
    """
    logger = get_logger()

    @wraps(func)
    def wrapper(*args, **kwargs):
        statement, engine = func(*args, **kwargs)
        time_started = time.time()
        records_size = kwargs['dataframe'].shape[0] if isinstance(
            kwargs['dataframe'], DataFrame) else len(kwargs['dataframe'])
        logger.debug('Method %s started ...' % func.__name__.title())
        result = engine.execute(
            statement
        )
        querying_time = time.time() - time_started
        logger.debug('Done Executing sql %s in %f seconds %s records' % (
            func.__name__, querying_time, records_size))
        return result

    return wrapper


class DataFrameExecuteSql:

    @classmethod
    @execute_sql
    def save_dataframe_to_sql_returning_id(cls, engine, dataframe, model,
                                           update_column, constraint):
        statement = cls.base_save_dataframe_to_sql(dataframe, model,
                                                   update_column, constraint)
        statement = statement.returning(model.id)
        return statement, engine

    @classmethod
    @execute_sql
    def save_dataframe_to_sql(cls, engine, dataframe, model, update_column,
                              constraint):
        statement = cls.base_save_dataframe_to_sql(dataframe, model,
                                                   update_column, constraint)
        return statement, engine

    @classmethod
    @execute_sql
    def save_dataframe_to_sqlite(cls, engine, dataframe, model,
                                 update_column, constraint):
        statement, record_columns = cls.get_insert_statement(model,
                                                             dataframe)
        return statement, engine

    @classmethod
    def base_save_dataframe_to_sql(cls, dataframe, model, update_column,
                                   constraint):
        statement, record_columns = cls.get_insert_statement(model,
                                                             dataframe)
        set_update_attributes = cls.get_save_attributes(statement,
                                                        update_column, model,
                                                        record_columns)
        if not set_update_attributes:
            statement = statement.on_conflict_do_nothing(
                constraint=constraint,
            )
        else:
            statement = statement.on_conflict_do_update(
                constraint=constraint,
                set_=set_update_attributes
            )
        return statement

    @classmethod
    @execute_sql
    def insert_dataframe_to_sql(cls, engine, dataframe, model,
                                update_column=None):
        statement, record_columns = cls.get_insert_statement(model,
                                                             dataframe)
        return statement, engine

    @classmethod
    def get_save_attributes(cls, statement, update_column, model,
                            record_columns=None):
        if isinstance(update_column, str):
            return {update_column: getattr(statement.excluded, update_column)}
        elif isinstance(update_column, tuple) or isinstance(update_column,
                                                            list):
            save_columns = {column: getattr(statement.excluded, column) for
                            column in update_column}
            if record_columns:
                save_columns = {column: value for column, value in
                                save_columns.items()
                                if column in record_columns}
            return save_columns
        else:
            raise AlchemyUpdateColumnsNotSupported(model)

    @classmethod
    @execute_sql
    def insert_dataframe_to_sql(cls, engine, dataframe, model,
                                update_column=None):
        statement, record_columns = cls.get_insert_statement(model,
                                                             dataframe)
        return statement, engine

    @classmethod
    @execute_sql
    def insert_dataframe_to_sql_returning_ids(cls, engine, dataframe, model):
        statement, record_columns = cls.get_insert_statement(model,
                                                             dataframe)
        statement = statement.returning(model.id)
        return statement, engine

    @classmethod
    def get_insert_statement(cls, model, dataframe):
        records_rows = cls.get_records(dataframe)
        statement = cls.prepare_bulk_statement(model, records_rows)
        return statement, records_rows[0].keys()

    @classmethod
    def get_records(cls, dataframe):
        if isinstance(dataframe, DataFrame) or hasattr(dataframe, 'to_dict'):
            return dataframe.to_dict('records')
        elif isinstance(dataframe, dict):
            return [dataframe]
        assert isinstance(dataframe, list)
        return dataframe

    @classmethod
    def prepare_bulk_statement(cls, model, records_rows):
        return insert(model.__table__).values(records_rows)


from pandas_orm.sqlalchemy.crud.save import ModelDataFrameManager
from pandas_orm.sqlalchemy.query import to_dataframe
from pandas_orm.sqlalchemy.dataframe import DataFrame, initialized_dataframe
from pandas_orm.sqlalchemy.session.sqlalchemy_db import DatabaseSession


class ModelManager(DatabaseSession):

    def __init__(self, model, *args, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        self.model = model

    @to_dataframe
    def all(self):
        with self.session() as session:
            return session.query(self.model)

    def delete(self, **kwargs):
        with self.session() as session:
            prepared_statement = session.query(self.model)
            for key, value in kwargs.items():
                prepared_statement = prepared_statement.filter(
                    getattr(self.model, key) == value
                )
            return prepared_statement.delete()

    @to_dataframe
    def get(self, **kwargs):
        with self.session() as session:
            prepared_statement = session.query(self.model)
            for key, value in kwargs.items():
                prepared_statement = prepared_statement.filter(
                    getattr(self.model, key) == value
                )
            return prepared_statement

    def bulk_save(self, records, unique_fields=None, update_fields=None, returning_id=False, model=None) -> DataFrame:
        model = model if model else self.model
        new_records = ModelDataFrameManager.bulk_save(
            records=records,
            model=model,
            unique_fields=unique_fields,
            update_fields=update_fields,
            returning_id=returning_id,
            engine_context_func=self.engine
        )
        return initialized_dataframe(new_records, self)









from django.db.models import Model
from pandas import DataFrame as PDDataFrame
from pandas_orm.base.dataframe import BaseDataFrame
from pandas_orm.base.exceptions import DataFrameModelNotSpecified


class DataFrame(BaseDataFrame):
    __model__: Model = None
    """
        declare
        :param __model__: pandas_orm.django.Model
    """

    def __init__(self, *args, orm_model=None, **kwargs):
        """
        :param args: default pandas.DataFrame args
        :param orm_model: pandas_orm.django.Model or from django.db.models import Model
        :param kwargs: default pandas.DataFrame kwargs
        """
        super(__class__, self).__init__(*args, **kwargs)
        self.__model__: Model = orm_model

    def to_objects(self):
        return [self.model(**record) for record in self.to_dict('records')]

    def bulk_update(self, fields=None, batch_size=None, model=None):
        """
        :param fields: List[str] fields to update for each records in the table
        :param batch_size: integer
        :param model: pandas_orm.django.Model
        :return:
        """
        model = self.validate_dataframe_orm(model)
        if self.empty:
            return self

        return self.model.objects.bulk_update(self, fields, batch_size,
                                              model=model)

    def bulk_create(self, *args, model=None, **kwargs):
        """
        :param args: extended bulk_create args
        :param kwargs: extended bulk_create kwargs
        :param model: pandas_orm.django.Model
        :return:
        """
        model = self.validate_dataframe_orm(model)
        if self.empty:
            return self
        return model.objects.bulk_create(self, *args, **kwargs)

    def validate_dataframe_orm(self, model=None):
        model = model if model else self.model
        if not model.objects.__class__.__name__ == 'DataFrameManager':
            raise DataFrameModelNotSpecified()
        return model


def is_dataframe(records):
    """
    :param records:
    :return: is records of instance DataFrame
    """
    return isinstance(records, DataFrame) or isinstance(records, PDDataFrame)



from django.db import models
from django.db.models import QuerySet
from django.db.models.manager import BaseManager

from pandas_orm.django.crud.save import bulk_create, bulk_update
from pandas_orm.django.query import to_dataframe
from pandas_orm.django.query import django_dataframe_values

DJANGO_SUPPORTED_FIELDS = '_state'


class DjangoDFQuerySet(QuerySet):
    @to_dataframe
    def to_dataframe(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return: DataFrame representation for the queryset
        """
        return self

    @django_dataframe_values
    def to_dataframe_values(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return: DataFrame representation for the queryset as values query
        """
        return self


class DataFrameManager(BaseManager.from_queryset(DjangoDFQuerySet)):

    def bulk_create(self, objs, *args, **kwargs):
        """
        bulk_create with DataFrame
        :param objs: List[django.db.models.Model] or DataFrame with similar model as dictionary
        :param args: extended bulk_create args
        :param kwargs: extended bulk_create kwargs
        :return:
        """
        return bulk_create(
            objs,
            model=self.model,
            func=super(__class__, self).bulk_create,
            args=args, kwargs=kwargs
        )

    def bulk_update(self, objs, *args, model=None, **kwargs):
        """
        bulk_update with DataFrame
        :param objs: List[django.db.models.Model] or DataFrame with similar model as dictionary
        :param args: extended bulk_update args
        :param kwargs: extended bulk_update kwargs
        :return:
        """
        return bulk_update(
            objs,
            model=self.model or model,
            func=super(__class__, self).bulk_update,
            args=args, kwargs=kwargs
        )


class DjangoDFModel(models.Model):
    objects = DataFrameManager()

    def to_dict(self):
        """
        :return: Dict representation
        """
        return {
            key: value
            for key, value in
            self.__dict__.items()
            if key not in DJANGO_SUPPORTED_FIELDS
        }

    class Meta:
        abstract = True



from pandas import DataFrame as PDDataFrame
from sqlalchemy.orm import declarative_base

from pandas_orm.base.dataframe import BaseDataFrame
from pandas_orm.sqlalchemy.crud.save import ModelDataFrameManager
from pandas_orm.sqlalchemy.exceptions import ModelIsMissing, \
    EngineScopeMissing


def is_dataframe(records):
    """
    :param records:
    :return: is the records of instance dataframe
    """
    return isinstance(records, DataFrame) or isinstance(records, PDDataFrame)


class DataFrame(BaseDataFrame):

    def __init__(self, *args, orm_model=None, model_manager=None, **kwargs):
        """
        :param args: default pandas.DataFrame args
        :param orm_model: declarative_base
        :param model_manager: pandas_orm.sqlalchemy.model_manager.ModelManager not required
        :param kwargs: default pandas.DataFrame kwargs
        """
        super(__class__, self).__init__(*args, **kwargs)
        self.__model__: declarative_base = orm_model
        self.model_manager = model_manager

    def get_model(self, model=None):
        if model:
            return model
        if self.model_manager:
            return self.model_manager.model
        return self.model

    def bulk_save(self, unique_fields=None, update_fields=None,
                  returning_id=False, engine_context_func=None, model=None):
        """
        :param unique_fields: unique List[field] of the table
        :param update_fields: update fields List[field] of the table
        :param returning_id: is returning id
        :return: pandas_orm.sqlalchemy.dataframe.DataFrame
        """
        model = self.get_model(model)
        if self.model_manager is None:
            if engine_context_func is None:
                raise EngineScopeMissing()
            if model is None:
                raise ModelIsMissing()
            df = ModelDataFrameManager.bulk_save(
                self,
                engine_context_func=engine_context_func,
                model=model,
                unique_fields=unique_fields,
                update_fields=update_fields,
                returning_id=returning_id,
            )
            return initialized_dataframe_model(df, model)
        return self.model_manager.bulk_create(
            self,
            unique_fields=unique_fields,
            update_fields=update_fields,
            returning_id=returning_id
        )

    def to_objects(self, model=None):
        model = model if model else self.model
        return [model(**record) for record in self.to_dict('records')]


def initialized_dataframe(records, model_manager=None) -> DataFrame:
    """
    :param records: collection or DataFrame
    :param model_manager: ModelManager
    :return: pandas_orm.sqlalchemy.dataframe.DataFrame
    """
    if isinstance(records, list):
        records = DataFrame(records, orm_model=model_manager.model, model_manager=model_manager)
    if not is_dataframe(records):
        raise NotImplementedError(type(records))

    df = DataFrame(records, orm_model=model_manager.model, model_manager=model_manager)
    return df


def initialized_dataframe_model(records, model) -> DataFrame:
    """
    :param records: collection or DataFrame
    :param model: ModelManager
    :return: pandas_orm.sqlalchemy.dataframe.DataFrame
    """
    if isinstance(records, list):
        records = DataFrame(records)
    if not is_dataframe(records):
        raise NotImplementedError(type(records))
    df = DataFrame(records, orm_model=model)
    return df



from pandas_orm.base.exceptions import PandasORMError


class AlchemyDataBaseError(PandasORMError):
    pass


class AlchemySaveModelStatementError(AlchemyDataBaseError):
    pass


class AlchemyModelManagerNotInitialized(AlchemyDataBaseError):
    pass


class AlchemyUpdateColumnsNotSupported(AlchemyDataBaseError):
    pass


class ToDataFrameNotEmpty(AlchemyDataBaseError):
    pass


class EngineScopeMissing(AlchemyDataBaseError):
    pass


class ModelIsMissing(AlchemyDataBaseError):
    pass



from typing import List
from sqlalchemy import UniqueConstraint


class NativeModelSaveArguments:

    def __init__(self, records, model, unique_fields=None, update_fields=None):
        self.model = model
        self.records = records
        self.unique_fields = unique_fields
        self.update_fields = update_fields

    def get_args(self) -> dict:
        constraints = self.get_unique_constraints()
        constraints_names = self.get_constraints_names(constraints)
        constraint = constraints_names[0] if constraints_names else None
        constraints_values = self.get_constrain_fields(constraints)
        update_columns = self.get_update_fields(constraints_values)
        kwargs = dict(
            dataframe=self.records,
            model=self.model,
            update_column=update_columns,
            constraint=constraint)
        return kwargs

    def get_constrain_fields(self, constraints=None):
        if self.unique_fields:
            return self.update_fields
        return [col for con in constraints for col in
                con._pending_colargs]

    def get_constraints_names(self, constraints):
        return [value.name for value in constraints]

    def get_unique_constraints(self):
        constraints = [value for value in self.model.__dict__.values() if
                       isinstance(value, UniqueConstraint)]
        return constraints

    def get_implementation_columns(self, models, exclude_columns):
        columns = []
        for table in models:
            columns += [
                col for col in table.__table__.columns if
                col.name not in exclude_columns
            ]
        return columns

    def get_update_fields(self, unique_columns: List[str]):
        columns = []
        if self.update_fields:
            return self.update_fields
        for column in self.model.__table__.columns:
            if column.primary_key:
                continue
            if column.unique:
                continue
            if column.name in unique_columns:
                continue
            columns.append(column.name)

        return columns



from pandas_orm.django.crud.naive_save_arguments import \
    NativeBulkCreateArguments
from pandas_orm.django.crud.naive_save_arguments import \
    NativeBulkUpdateArguments
from pandas_orm.django.dataframe import is_dataframe
from pandas_orm.django.mixins.dataframe_mixin import DjangoDataFrameMixin


def bulk_create(dataframe, model, func, args, kwargs):
    """
    :param dataframe: List[django.db.models.Model] or DataFrame with similar model as dictionary
    :param model: django.db.models.Model
    :param func: Model.objects.bulk_create or ModelManager func
    :param args: func args
    :param kwargs: func kwargs
    :return: created DataFrame
    """
    if not is_dataframe(dataframe):
        return func(dataframe, *args, **kwargs)
    if dataframe.empty:
        return dataframe

    records = DjangoDataFrameMixin.get_objs(dataframe, model)
    naive_args = NativeBulkCreateArguments(model=model, dataframe=dataframe)
    nargs, nkwargs = naive_args.get_args(*args, **kwargs)
    saved = func(records, *nargs, **nkwargs)
    DjangoDataFrameMixin.add_dataframe_ids(dataframe, saved)
    return dataframe


def bulk_update(dataframe, model, func, args, kwargs):
    """
    :param dataframe: List[django.db.models.Model] or DataFrame with similar model as dictionary
    :param model: django.db.models.Model
    :param func: Model.objects.bulk_update or ModelManager func
    :param args: func args
    :param kwargs: func kwargs
    :return: updated DataFrame
    """
    if not is_dataframe(dataframe):
        return func(dataframe, *args, **kwargs)
    if dataframe.empty:
        return dataframe

    records = DjangoDataFrameMixin.get_objs(dataframe, model)
    naive_args = NativeBulkUpdateArguments(model=model, dataframe=dataframe)
    nargs, nkwargs = naive_args.get_args(*args, **kwargs)
    saved = func(records, *nargs, **nkwargs)
    DjangoDataFrameMixin.add_dataframe_ids(dataframe, saved)
    return dataframe



from django.db.models import Model


class NaiveModelBulkArguments:
    def __init__(self, model: Model, dataframe):
        self.model = model
        self.records = dataframe

    def get_update_fields(self, fields=None):
        if fields:
            return fields
        """
        update_fields=['last_name'], unique_fields=['name', 'email']
        """
        columns = []
        for column in self.model._meta.fields:
            if column.primary_key:
                continue
            if column.unique:
                continue
            columns.append(column.name)
        return columns


class NativeBulkCreateArguments(NaiveModelBulkArguments):

    def get_args(self,
                 batch_size=None,
                 ignore_conflicts=None,
                 update_conflicts=None,
                 update_fields=None,
                 unique_fields=None, *args, **kwargs
                 ):
        naive_ignore_conflicts = self.get_ignore_conflicts(ignore_conflicts)
        naive_update_conflicts = self.get_update_conflicts(update_conflicts)
        naive_update_fields = self.get_update_fields(update_fields)
        naive_unique_fields = self.get_unique_fields(unique_fields)
        naive_kwargs = dict(
            ignore_conflicts=naive_ignore_conflicts,
            update_conflicts=naive_update_conflicts,
            update_fields=naive_update_fields,
            unique_fields=naive_unique_fields,
            batch_size=batch_size,
        )
        return (), {**naive_kwargs, **kwargs}

    def get_unique_fields(self, fields):
        if fields:
            return fields
        if self.model._meta.constraints:
            return self.model._meta.constraints[0].fields

        return []

    def get_update_conflicts(self, is_conflict):
        if is_conflict is not None:
            return is_conflict

        return True

    def get_ignore_conflicts(self, is_conflict):
        if is_conflict is not None:
            return is_conflict

        return False


class NativeBulkUpdateArguments(NaiveModelBulkArguments):

    def get_args(self, fields=None, batch_size=None):
        naive_fields = self.get_fields(fields)
        naive_kwargs = dict(
            fields=naive_fields,
            batch_size=batch_size,
        )
        return (), naive_kwargs

    def get_fields(self, fields):
        return self.get_update_fields(fields)



import time
from functools import wraps

import sqlalchemy
from pandas import read_sql, concat
from pandas import DataFrame as PDDataFrame
from sqlalchemy.orm import Query

from pandas_orm.base.log import get_logger
from pandas_orm.sqlalchemy.exceptions import ToDataFrameNotEmpty

from pandas_orm.sqlalchemy.dataframe import DataFrame
from pandas_orm.sqlalchemy.dataframe import is_dataframe


def to_dataframe_gen(func):
    """
    Annotation that wraps function that returns generator(sqlalchemy.orm.query.Query)
     reads it to a dataframe
    Reads the query statements as a dataframe
    :param func: function that returns generator/list sqlalchemy.Session.query
    :return: DataFrame
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        gen = func(*args, **kwargs)
        dataframes = []
        for result_query in gen:
            df = query_to_dataframe(result_query)
            dataframes.append(df)
        df_results = concat(dataframes)
        return df_results

    return wrapper


def initialize_queryset_dataframe(df: PDDataFrame, query: Query):
    if len(query.column_descriptions) == 1:
        column_description = query.column_descriptions[0]
        if column_description['aliased'] is False:
            return DataFrame(df, orm_model=column_description['entity'])
    return DataFrame(df)


def to_dataframe(func):
    """
    Annotation that wraps function that returns sqlalchemy.orm.query.Query
     reads it to a dataframe
    :param func: function that returns sqlalchemy.Session.query
    :return: DataFrame
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        query = func(*args, **kwargs)
        df = query_to_dataframe(query)
        wrapped_df = initialize_queryset_dataframe(df, query)
        if args and args[0].__class__.__name__ == 'ModelManager':
            wrapped_df.model_manager = args[0]
            wrapped_df.__model__ = wrapped_df.model_manager.model
        return wrapped_df

    return wrapper


def query_to_dataframe(query: sqlalchemy.orm.query.Query):
    """
    Return a DataFrame read from sqlalchemy.orm.query.Query
    :param query: : sqlalchemy.orm.query.Query
    :return: dataframe represent of the result
    :rtype: DataFrame
    """
    logger = get_logger()
    if query is None:
        return DataFrame([])
    if is_dataframe(query):
        if not query.empty:
            raise ToDataFrameNotEmpty(query)
        return query
    assert isinstance(query, Query)
    time_started = time.time()
    logger.debug('Querying started %s ...' % query.statement)
    result_dataframe = read_sql(query.statement,
                                query.session.bind)
    querying_time = time.time() - time_started
    logger.debug('Done Querying in %f seconds ' % querying_time)
    return result_dataframe



from django.db import models as django_models

from pandas_orm.django.model import DjangoDFModel as Model
from pandas_orm.django.model import DataFrameManager

models = django_models
models.Model = Model
models.manager.BaseManager = DataFrameManager